/*
%	VFFTFILTER . C
%
%	Circular Filter for VFFT image transform.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

This software is copyright (C) by the Lawrence Berkeley Laboratory.
Permission is granted to reproduce this software for non-commercial
purposes provided that this notice is left intact.

It is acknowledged that the U.S. Government has rights to this software
under Contract DE-AC03-765F00098 between the U.S.  Department of Energy
and the University of California.

This software is provided as a professional and academic contribution
for joint exchange. Thus, it is experimental, and is provided ``as is'',
with no warranties of any kind whatsoever, no support, no promise of
updates, or printed documentation. By using this software, you
acknowledge that the Lawrence Berkeley Laboratory and Regents of the
University of California shall have no liability with respect to the
infringement of other copyrights by any part of this software.

For further information about this notice, contact William Johnston,
Bld. 50B, Rm. 2239, Lawrence Berkeley Laboratory, Berkeley, CA, 94720.
(wejohnston@lbl.gov)

For further information about this software, contact:
	Jin Guojun
	Bld. 50B, Rm. 2275, Lawrence Berkeley Laboratory, Berkeley, CA, 94720.
	g_jin@lbl.gov

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% compile:
%	cc -O -o DEST/vfftfilter vfftfilter.c -lscs2 -lccs -lhips -lrle -lm
*/
char	usage[]="options:\n\
[-base #]	base value. [Default=0.0], good max => 2.0\n\
[-c]		convex curve. Default is concave curve\n\
[-hi]		high-pass filter. Default is low-pass\n\
[-fe #]		elastic scale factor\n\
[-func #]	filter function\n\
[-F #]		number of Frames in sample [1 in 2D]. more F # is 3D\n\
[-G]		generate gnu_plot data of curves only\n\
[-L]		Linear filter\n\
[-min #]	minimum value at bottom [0.01]\n\
[-power #]	powers [2]\n\
[-r #]		radius [0.333]\n\
[-R #]		radius factor. default is 1 (100%). Range 0.1 - 1.0\n\
[-v]		verbose\n\
[<] image [> [-o] VFFT]\n\n\
filter functions:	\n\
default:	elastic	\n\
1		ideal	\n\
2		exponential\n\
3		butterworth\n\
4		right tri-angle\n\
5		user defined standard table\n\
6 table_name	user defined table (ascii)\n\
7 table_name	user defined table (binary)\n",
fname[128];
/*
%
% AUTHOR:	Jin Guojun - Lawrence Berkeley Laboratory	5/1/91
*/

#include "vfft_fil.h"

U_IMAGE	uimg;

bool	dimens, curve_dir, hi_pass, Table, Msg;
VType	*itemp;
Filter	*lkt, flr=0.01, power=2.0, radius=.33, sc;
int	dimen1len, h_rows, h_frm, sframes=1,
	r_radius, c_radius, f_radius, max_radius;
MType	fsize, vsize;


main (argc, argv)
int	argc;
char**	argv;
{
bool	dflag=0, vflag, sample=0;
int	f, fn, j, max_dis;
MType	*ldis_p2, *d2_dis_p2;
Filter	base=0, radius_rate=1., *filter3d, **mask_plane;
register int	i;

format_init(&uimg, IMAGE_INIT_TYPE, HIPS, *argv, "S20-1");

for (j=1; j<argc; j++)
    if (*argv[j] == '-'){
	f=1;
	switch (argv[j][f++]){
	case 'h':
		hi_pass++;	break;
	case 'L':
		curve_dir=CURVE_LINEAR;	break;
	case 'c':
		curve_dir=CURVE_CONVEX;	break;
	case 'b':
		if (GValue(0))
			base = atof(argv[j]+f);
		break;
	case 'f':{
	register int	who = argv[j][f];
		if (GValue(0))
		   if (who == 'u'){
			fn = atoi(argv[j]+f);
			if (fn > 5 && j+1<argc)
				strcpy(fname, argv[++j]);
		   }
		   else	sc = atof(argv[j]+f);
	}	break;
	case 'F':
		if (GValue(0))
			sframes = atoi(argv[j]+f);
		break;
	case 'G':
		Table++;	break;
	case 'm':
		if (GValue(0))
			flr = atof(argv[j]+f);
		break;
	case 'p':
		if (GValue(0))
			power = atof(argv[j]+f);
		break;
	case 'r':
		if (GValue(0))
			radius = atof(argv[j]+f);
		break;
	case 'R':
		if (GValue(0))
			radius_rate = atof(argv[j]+f);
		if (radius_rate < 0 || radius_rate > 1.0)
			radius_rate = 1.0;
		break;
	case 'v':	Msg++;
		break;
#ifdef	IBMPC
	case 'i':if (GValue(1) &&
		    !(in_fp=freopen(argv[j]+f, "rb", stdin)))
			syserr("can't open %s as image input", argv[j]);
		break;
#endif
	case 'o':if (GValue(1) &&
		    !freopen(argv[j]+f, "wb", stdout))
			syserr("%s can't be opened", argv[j]);
		break;
	default:
info:		usage_n_options(usage, j, argv[j]);
	}
    }
    else if ((in_fp=freopen(argv[j], "r", stdin)) == NULL)
	syserr("can't open frame file - %s", argv[j]);

io_test(stdout_fd, goto info);

io_test(fileno(in_fp), sample=!iset);

if (sample){	/* no input file */
	cln = row = 128;
	uimg.frames = sframes + !sframes;
	uimg.pxl_in = 8;
	uimg.in_form = IFMT_VFFT3D;
	(*uimg.std_swif)(FI_INIT_NAME, &uimg, *argv, 0);
}
else	{
	(*uimg.header_handle)(HEADER_READ, &uimg, 0, 0);
	if (uimg.in_form < IFMT_VFFT3D || uimg.in_form  > IFMT_VVFFT3D)
		syserr("image is not a VFFT format");
	vflag = uimg.in_form==IFMT_VVFFT3D || uimg.in_form==IFMT_DVVFFT3D;
}
uimg.o_form = uimg.in_form;
uimg.pxl_out = uimg.pxl_in;

if (!vflag)
	dimens = !(uimg.in_form & 1); /* if it is 2D */
dimen1len = (cln>>1) + 1;
h_rows = row >> 1;
h_frm = frm >> 1;
fsize = row*cln;
vsize = row*dimen1len;
f_radius = radius_rate * h_frm;
r_radius = radius_rate * h_rows;
c_radius = radius_rate * dimen1len;
if (f_radius < 1)	f_radius = 1;
if (r_radius < 1)	r_radius = 1;
if (c_radius < 1)	c_radius = 1;

max_radius = MAX(MAX(h_frm, h_rows), dimen1len);
max_dis = sqrt((double)(h_frm*h_frm + h_rows*h_rows + dimen1len*dimen1len));
lkt = zalloc(max_dis, sizeof(*lkt), "lkt");
ldis_p2 = zalloc(max_radius, sizeof(*ldis_p2), "ldis_p2");
d2_dis_p2 = nzalloc((max_radius+1)*max_radius>>1, sizeof(*d2_dis_p2), "d2d");

filter3d = nzalloc(sizeof(*filter3d), max_radius*max_radius, "filter3d");
ibuf = nzalloc(uimg.pxl_in, vsize, "ibuf");
obuf = zalloc(uimg.pxl_in, vsize, "obuf");	/* be zeroed */

mask_plane = nzalloc(max_radius, sizeof(*mask_plane), "mask_p");
mask_plane[0] = filter3d;
for (i=1; i<max_radius; i++)
	mask_plane[i] = mask_plane[i-1] + max_radius;

#ifdef	_DEBUG_
message("Max_R=%d, f_r=%d, r_r=%d, c_r=%d, max_dis=%d, func=%d\n",
	max_radius, f_radius, r_radius, c_radius, max_dis, fn);
#endif

/*	filling distance line	*/
for (i=1; i<max_radius; i++)
	ldis_p2[i] = i*i;
/*	filling distance right tri_angle plane	*/
{
register int	r, c, *dp=d2_dis_p2;
for (r=0; r<max_radius; r++)
    for (c=r; c<max_radius; c++)
	*dp++ = ldis_p2[r] + ldis_p2[c];
}

if (fn>1 && sc)	radius = sc * 0.002;
if (hi_pass)	sc = -sc;
lkt_function(lkt, fn, max_dis);

if (Table)	exit(0);	/* only send plot data to stdout. */

(*uimg.header_handle)(HEADER_WRITE, &uimg, argc, argv, True);

if (sample){
	{
	register float	*ibp=ibuf;
	    for (i=vsize << 1; i--;)
		ibp[i] = 256.;
	}
	for (f=h_frm; f--;){
		fill_mask_plane(d2_dis_p2, max_radius, ldis_p2[f], mask_plane, lkt);
		filtering(ibuf, obuf, base, filter3d, f+1);
	}
	if (frm & 1){
		fill_mask_plane(d2_dis_p2, max_radius, 1, mask_plane, lkt);
		filtering(ibuf, obuf, base, filter3d, f+1);
	}
	for (f=0; f<h_frm; f++){
		fill_mask_plane(d2_dis_p2, max_radius, ldis_p2[f], mask_plane, lkt);
		filtering(ibuf, obuf, base, filter3d, f+h_frm);
	}
	exit(0);
}

if (dimens){
	message("%s: 2 dimension filter - %d frames\n", Progname, frm);
	fill_mask_plane(d2_dis_p2, max_radius, ldis_p2[0], mask_plane, lkt);
	for (f=0; f<frm; f++){
		i = upread(ibuf, uimg.pxl_in, vsize, in_fp);
		if (i != vsize)
			syserr("f%d [%d] read %d", f+1, vsize, i);
		filtering(ibuf, obuf, base, filter3d, f+1);
	}
}
else	{
	for (f=0; f<f_radius; f++){
		fill_mask_plane(d2_dis_p2, max_radius, ldis_p2[f], mask_plane, lkt);
		i = upread(ibuf, uimg.pxl_in, vsize, in_fp);
		if (i != vsize)
			syserr("f%d [%d] read %d", f+1, vsize, i);
		filtering(ibuf, obuf, base, filter3d, f+1);
	}

	for (; frm>1 && f<frm-(f_radius<<1); f++){	/* maybe send all 0 frames */
		fill_mask_plane(d2_dis_p2, max_radius, ldis_p2[f_radius], mask_plane, lkt);
		i = upread(ibuf, uimg.pxl_in, vsize, in_fp);
		if (i != vsize)
			syserr("f%d [%d] read %d", f+1, vsize, i);
		filtering(ibuf, obuf, base, filter3d, f+1);
	}

	for (; f<frm; f++){
		fill_mask_plane(d2_dis_p2, max_radius, ldis_p2[frm-1-f], mask_plane, lkt);
		i = upread(ibuf, uimg.pxl_in, vsize, in_fp);
		if (i != vsize)
			syserr("f%d [%d] read %d", f+1, vsize, i);
		filtering(ibuf, obuf, base, filter3d, f+1);
	}

}
}

new_curve(lkt, maxdiff, neg, curve, name)
float*	lkt;
int	maxdiff;
char	*name;
{
register int	i;
Filter	rel_val, maxout=1.;
register Filter	scale, vsc=.01*sc/maxdiff + 1;

if (maxdiff < 2){
	lkt[0] = 1.;
	return;
}
if (curve==CURVE_LINEAR){
	scale = (maxout-flr)/maxdiff;
	if (neg) for (i=maxdiff; i--;)
		lkt[i] = i * scale + flr;
	else for (i=maxdiff, maxdiff--; i--;)
		lkt[maxdiff-i] = i * scale + flr;
}
else{
	{
	register double	tmp;
	    if (vsc != 1.){
		for (i=tmp=maxdiff; i--;)	tmp = tmp*vsc + i;
		scale = (maxout-flr) / tmp;
	    }
	else	scale = (maxout-flr) / ((maxdiff-1)*maxdiff >> 1);
	}
	if (!neg)
	   if (!curve)
		for(i=maxdiff, rel_val=flr; i--;)
		{
		rel_val += (maxdiff-i) * (scale*=vsc);
		if (rel_val > maxout)	lkt[i]=maxout;
		else	lkt[i] = rel_val;
		}
	   else for (i=0, rel_val=maxout; i<maxdiff; i++)
		{
		lkt[i] = rel_val;
		rel_val -= i * (scale*=vsc);
		if (rel_val<flr && !uimg.in_form)	rel_val=flr;
		}
	else if (!curve)
		for(i=0, rel_val=flr; i<maxdiff; i++)
		{
		rel_val += i * (scale*=vsc);
		lkt[i] = rel_val;
		}
	   else for (i=maxdiff, rel_val=maxout; i--;)
		{
		lkt[i] = rel_val;
		rel_val -= (maxdiff-i) * (scale*=vsc);
		}
	}
if (Msg){
	dump_lkt(lkt, maxdiff);
	msg("\n%d %s => sc=%f, scale=%f\n", maxdiff, name, sc, scale);
}
}

dump_lkt(lktp, n)
register float	*lktp;
{
register int	i;
for (i=0; i<n;){
	message("%10.3f", lktp[i]);	if (!(++i & 7))	mesg("\n");	}
}

GTable(lp, max)
register float *lp;
{
register int i;
for (i=0; i<max; i++)
	printf("%d %f\n", i, *lp++);
printf("\n\n");
}

filtering(i_buf, o_buf, base_v, filterp, f)
VType	*i_buf, *o_buf;
Filter	base_v;
register Filter	*filterp;
{
register float	*ibp=i_buf, *obp=o_buf;
register int	i, j;

	if (base_v)
	    for (i=max_radius*max_radius; i--;)
		filterp[i] += base_v;
	ibp += hi_pass * ((h_rows-r_radius+1)*dimen1len - c_radius);

	for (i=0; i<r_radius; i++){
	register int	dis = (row - 1 - (i<<1)) * dimen1len << 1;
	    for (j=0; j<c_radius; j++){
		obp[dis] = ibp[dis] * *filterp;
		*obp++ = *ibp++ * *filterp;
		obp[dis] = ibp[dis] * *filterp;
		*obp++ = *ibp++ * *filterp++;
	    }
	    obp += dimen1len - c_radius << 1;
	    ibp += dimen1len - c_radius << 1;
	    filterp += max_radius - c_radius;
	}

	i = fwrite(o_buf, uimg.pxl_in, vsize, out_fp);
	if (i != vsize)
		syserr("f%d [%d] write %d", f, vsize, i);
}

fill_mask_plane(dis_p2, ldis_len, framedis, filter_buf, lkt)
register MType	*dis_p2;
register Filter	**filter_buf, *lkt;
{
register int	r, c;
register Filter	fdis = framedis;

for (r=0; r<ldis_len; r++){
    filter_buf[r][r] = lkt[(int)(sqrt(fdis+*dis_p2++) + 0.5)];
    for (c=r+1; c<ldis_len; c++){
    register int distance = sqrt(fdis+*dis_p2++) + 0.5;
	filter_buf[r][c] = filter_buf[c][r] = lkt[distance];
    }
/*	filter_buf[r][c] = filter_buf[c][r] =
		lkt[(int)(sqrt(fdis+*dis_p2++) + 0.5)];
*/
}
}

Filter
power_func(fn, i, r)
register int	fn, i;
register Filter	r;
{
    if (fn == FUNC_EXP)
	if (curve_dir)	r = exp(-pow(r / i, power));
	else		r = exp(-pow(i / r, power));
    else if (curve_dir)	r = 1. / (1 + pow(r / i, power));
	else		r = 1. / (1 + pow(i / r, power));
return	r;
}

lkt_function(lkt, fn, ns)
Filter	*lkt;
{
char	str[128];
register Filter	R = radius * ns;
register int	i, rev;

sprintf(str, "radius = %.3f, filter is ", (fn<6) ? radius : 1.);

switch (fn) {
case FUNC_IDEAL:
	sprintf(fname, "ideal filter %.2f\n", R);
	strcat(str, fname);
	if (hi_pass)
	    for (i=R; i<ns; i++)
		lkt[i] = 1.;
	else
	    for (i=R; i--;)
		lkt[i] = 1.;
	break;
case FUNC_EXP:
case FUNC_BUTTERWORTH:
	strcat(str, fn==FUNC_EXP ? "exponetial\n" : "butterworth\n");
	if (hi_pass){
		lkt[0] = flr;
		if (curve_dir) for (i=ns; --i;)
			lkt[i] = power_func(fn, i, R);
		else for (i=ns, rev=ns-1; --i;)
			lkt[rev-i] = power_func(fn, i, R);
	}
	else if (curve_dir)	for (i=ns, rev=ns-1; i--;)
			lkt[rev-i] = power_func(fn, i, R);
	else	for (i=ns; i--;)
			lkt[i] = power_func(fn, i, R);
	break;
case 4:
	strcat(str, "right tri_angle\n");
	if (hi_pass)
	    for (i=0; i<ns; i++)
		lkt[i] = (i+flr) / ns;
	else
	    for (i=0; i<ns; i++)
		lkt[i] = (ns-1-i+flr) / ns;
	break;
case 5:
case 6:	case 7:
	if (strlen(fname)==0)
		strcpy(fname, "filter_table");
	{
	char	buf[64];
	int	len;
	FILE	*fp = fopen(fname, fn==FUNC_ASCII_TABLE ? "r" : "rb");
		if (!fp)	syserr("open %s", fname);
		if (fn==FUNC_ASCII_TABLE)	{
			fscanf(fp, "%d", &len);
			if (len > ns)	len = ns;
			for (i=0; i<len; i++)
				fscanf(fp, "%f", lkt+i);
		} else	{
			len = getw(fp);	/* get table length */
			if (len > ns)	len = ns;
			fread(lkt, sizeof(*lkt), len, fp);
		}
		sprintf(buf, "%d user defined %s\n", len, fname);
		strcat(str, buf);
		for (;i<ns; i++)
			lkt[i] = lkt[i-1];
	}
	break;
default:
	strcat(str, "elastic\n");
	new_curve(lkt, ns, hi_pass, curve_dir, "ela");
}
mesg(str);
if (Table)	GTable(lkt, ns);
(*uimg.header_handle)(ADD_DESC, &uimg, str);
}
